#include <drivers.h>
#include <stdarg.h>
#include <minix/ioctl.h>
#include <minix/endpoint.h>
#include <fcntl.h>
#include <ibm/pci.h>
#include <sys/svrctl.h>
#include <sys/types.h>

#include <minix/ds.h>
#include <minix/syslib.h>
#include <minix/vm.h>
#include <timers.h>
#include <stdio.h>

#include "minixdefs.h"
#include "devid.h"


extern int oss_major;

/* The PCIBIOS_* defines must match oss_pci.h */
#define PCIBIOS_SUCCESSFUL 0x00
#define PCIBIOS_FAILED	-1



/* defines for output to console */
#define LOG_UNKNOWN_MESSAGE 1
#define NO_LOG_UNKNOWN_MESSAGE 0
#define DEBUG_MESSAGES 0



#define MAX_DEVICES 16
#define MAX_INTRS 32



typedef struct {
	char *name;
	int *var;
} option_t;

#include "minix_options.h"



typedef struct _module_t {
	char *name;
	int (*attach)(oss_device_t *);
	device_id_t *id_table;
	int *type;
} module_t;

#include "oss_modules.h"



typedef struct {
	int irq;
	int irq_hook_id;
	oss_device_t *osdev;
	oss_tophalf_handler_t top;
	oss_bottomhalf_handler_t bottom;
} osscore_intr_t;

static osscore_intr_t intrs[MAX_INTRS] = { {0} };
static int nintrs = 0;


static char io_ctl_buf[_IOCPARM_MASK];
static uio_t uio_bufs[/*OSS_MAX_CDEVS*/256] = { {0} };





#define MAX_GLOBAL_MESSAGES 10
static message global_messages[MAX_GLOBAL_MESSAGES];
static int free_global_message_index = 0;
static int index_global_messages = 0;


static timer_t *timers;




extern void soundcard_attach(void);
extern oss_device_t *osdev_create(dev_info_t *dip, int dev_type, int instance, 
		const char *nick, const char *handle);
extern int oss_open(int dev, int flags, int proc_nr);
extern int oss_ioctl(int dev, int cmd, char *arg);
extern int oss_close(int dev);
extern int oss_read(int dev, uio_t *buf);
extern int oss_write(int dev, uio_t *buf);



int oss_hz = DEFAULT_HZ;



static dev_info_t* minix_devices[MAX_DEVICES];
static oss_device_t *device_list[MAX_DEVICES];
static int ndevs = 0;






void cmn_err (int level, char *s, ...) {
	char tmp[1024], *a[6];

	va_list ap;
	int i, n = 0;

	va_start (ap, s);

	for (i = 0; i < strlen (s); i++)
		if (s[i] == '%')
			n++;

	for (i = 0; i < n && i < 6; i++)
		a[i] = va_arg (ap, char *);

	for (i = n; i < 6; i++)
		a[i] = NULL;

	strcpy (tmp, "OSS: "); 
	sprintf (tmp + strlen (tmp), s, 
			a[0], a[1], a[2], a[3], a[4], a[5], NULL, NULL, NULL, NULL);

	if (level == CE_PANIC)
		panic ("OSS", tmp, NO_NUM);

	printf("%s", tmp);
#if 0
	/* This may cause a crash under SMP */
	if (sound_started)
		store_msg (tmp);
#endif

	va_end (ap);
}



void reply_with_grant(int type, int replyee, int rep_endpt, int status, 
		unsigned int grant, const char *s) {
	message mess;

	mess.m_type = type;
	mess.REP_STATUS = status;
	mess.REP_ENDPT = rep_endpt;
	mess.REP_IO_GRANT = (unsigned) grant;
	send(replyee, &mess);
#if DEBUG_MESSAGES
	cmn_err(0, "replied '%s'\n", s);
#endif
}




void set_timer(timer_t *timer, clock_t ticks, tmr_func_t watchdog, void *arg) {
	int r;
	clock_t now, prev_time = 0, next_time;

	if ((r = getuptime(&now)) != OK)
		cmn_err(CE_PANIC, "Can't get uptime\n");

	/* Set timer argument and add timer to the list. */
	tmr_arg(timer)->ta_ptr = arg;
	prev_time = tmrs_settimer(&timers, timer, now+ticks, watchdog, &next_time);

	/* Reschedule our synchronous alarm if necessary. */
	if (! prev_time || prev_time > next_time) {
		if (sys_setalarm(next_time, 1) != OK)
			cmn_err(CE_PANIC, "Can't set alarm\n");
	}
}



void expire_timers(void) {
	clock_t next_time;
	clock_t now;
	int s;

	if ((s= getuptime(&now)) != OK) {
		cmn_err(0, "Cannot get uptime\n");
	}

	/* Check for expired timers and possibly reschedule an alarm. */
	tmrs_exptimers(&timers, now, &next_time);
	if (next_time > 0) {
		if (sys_setalarm(next_time, 1) != OK)
			cmn_err(CE_PANIC, "Can't set alarm.");
	}
}



void cancel_timer(timer_t *timer) {
	clock_t next_time, prev_time;
	prev_time = tmrs_clrtimer(&timers, timer, &next_time);

	/* If the earliest timer has been removed, we have to set the alarm to the
	 * next timer, or cancel the alarm altogether if the last timer has been
	 * cancelled (next_time will be 0 then).
	 */
	if (prev_time < next_time || ! next_time) {
		if (sys_setalarm(next_time, 1) != OK)
			panic(__FILE__, "OSS expire timer couldn't set alarm.", NO_NUM);
	}
}


int add_to_message_queue(message mess) {
	if (free_global_message_index == MAX_GLOBAL_MESSAGES) {
		return -2;
	}
	global_messages[free_global_message_index] = mess;
	free_global_message_index++;
	return 0;
}


message get_message_from_queue(void) {
	int index;

	if (index_global_messages == free_global_message_index - 1) {
		index = index_global_messages;
		index_global_messages = 0;
		free_global_message_index = 0;

		return global_messages[index];
	}
	else {
		index_global_messages++;
		return global_messages[index_global_messages - 1];
	}
}


int messages_in_queue(void) {
	return free_global_message_index > 0;
}


void reply(int type, int replyee, int rep_endpt, int status, const char *s) {
	reply_with_grant(type, replyee, rep_endpt, status, 0, s);
}


void oss_udelay(unsigned int t) {
	micro_delay((u32_t)t);
}


int oss_uiomove(void *address, size_t nbytes, enum uio_rw rwflag, 
		uio_t *uio_p) {
	int result;

	if (rwflag == UIO_READ) {
		result = sys_safecopyto(uio_p->revive_proc_nr,
				(vir_bytes)uio_p->revive_grant, uio_p->revive_status,
				(vir_bytes)address,(phys_bytes)nbytes, D);
		uio_p->revive_status += nbytes;
	}
	else if (rwflag == UIO_WRITE) {
		result = sys_safecopyfrom(uio_p->revive_proc_nr,
				(vir_bytes)uio_p->revive_grant, uio_p->revive_status,
				(vir_bytes)address,(phys_bytes)nbytes, D);
		uio_p->revive_status += nbytes;
	}
	else {
		cmn_err(CE_PANIC, "oss_uiomove(), ERROR: unknown rwflag: %d\n", rwflag);
	}

	if (result != 0) {
		cmn_err(CE_WARN, "%s\n", "oss_uiomove(), ERROR: sys_safecopy* failed");
		return EFAULT;
	}
	return 0;
}


#if 0
void tell_dev(vir_bytes buf, size_t size, int pci_bus, 
		int pci_dev, int pci_func) {
	int r;
	endpoint_t dev_e;
	u32_t u32;
	message m;

	r= ds_retrieve_u32("amddev", &u32);
	if (r != OK)
	{
		printf("tell_dev: ds_retrieve_u32 failed for 'amddev': %d\n",
				r);
		return;
	}

	dev_e= u32;

	m.m_type= IOMMU_MAP;
	m.m2_i1= pci_bus;
	m.m2_i2= pci_dev;
	m.m2_i3= pci_func;
	m.m2_l1= buf;
	m.m2_l2= size;

	r= sendrec(dev_e, &m);
	if (r != OK)
	{
		printf("tell_dev: sendrec to %d failed: %d\n", dev_e, r);
		return;
	}
	if (m.m_type != OK)
	{
		printf("tell_dev: dma map request failed: %d\n", m.m_type);
		return;
	}
}
#endif


void *oss_contig_malloc(unsigned long buffsize, unsigned long memlimit,
		oss_native_word *phaddr) {
	char *tmpbuf;

	*phaddr = 0;

	tmpbuf = alloc_contig(buffsize, AC_ALIGN4K, phaddr);

	if (tmpbuf == NULL)
	{
		cmn_err(CE_WARN, "Cannot allocate %u bytes for a DMA buffer\n",
				buffsize);
		cmn_err(CE_WARN, "Run soundoff and run soundon again.");
		return NULL;
	}

#if 0
	tell_dev((vir_bytes)tmpbuf, buffsize, 0, 0, 0);
#endif

	return tmpbuf;
}


void oss_unregister_interrupts(oss_device_t *osdev) {
	cmn_err(0, "Not implemented: oss_unregister_interrupts()");
	/* TODO */
}


int oss_register_interrupts(oss_device_t *osdev, int intrnum,
		oss_tophalf_handler_t top, oss_bottomhalf_handler_t bottom) {
	int retval;
	osscore_intr_t *intr;

	if (nintrs >= MAX_INTRS) {
		cmn_err(0, "Too many interrupt handlers");
		return -ENOMEM;
	}

	intr = &intrs[nintrs];

	intr->irq = osdev->dip->irq;
	intr->irq_hook_id = nintrs;
	intr->osdev = osdev;
	intr->top = top;
	intr->bottom = bottom;



	if ((retval = sys_irqsetpolicy(osdev->dip->irq, /*IRQ_REENABLE*/0,
					&(intr->irq_hook_id))) != OK) {
		cmn_err(CE_PANIC, "Cannot set irq policy, return code: %d\n", retval);
	}

	if ((retval = sys_irqenable(&(intr->irq_hook_id))) != OK) {
		cmn_err(CE_PANIC, "Cannot enable irq, return code: %d\n", retval);
	}

	nintrs++;

	return 0;
}


u32_t pci_read_config(dev_info_t *dip, offset_t where, int bytes) {
	switch (bytes) {
		case 1: return (u32_t) pci_attr_r8(dip->devind, (U16_t) where);
		case 2: return (u32_t) pci_attr_r16(dip->devind, (U16_t) where);
		case 4: return (u32_t) pci_attr_r32(dip->devind, (U16_t) where);
		default:
				cmn_err(CE_PANIC, "Wrong call to pci_read_config()");
				return (u32_t) -1;
	}
}


void pci_write_config(dev_info_t *dip, offset_t where, unsigned int val,
		int bytes) {

	switch (bytes) {
		case 1: pci_attr_w8(dip->devind, (U16_t) where, (U8_t) val);
				break;
		case 2: pci_attr_w16(dip->devind, (U16_t) where, (U16_t) val);
				break;
		case 4: pci_attr_w32(dip->devind, (U16_t) where, (U32_t) val);
				break;
		default:
				cmn_err(CE_PANIC, "Wrong call to pci_write_config()");
	}
}


int pci_read_config_byte(oss_device_t * osdev, offset_t where,
		unsigned char *val) {
	*val = (unsigned char) pci_read_config(osdev->dip, where, 1);

	return PCIBIOS_SUCCESSFUL;
}


int pci_read_config_irq(oss_device_t * osdev, offset_t where,
		unsigned char *val) {
	*val = (unsigned char) pci_read_config(osdev->dip, where, 1);

	return PCIBIOS_SUCCESSFUL;
}


int pci_read_config_word(oss_device_t *osdev, offset_t where, 
		unsigned short *val) {
	*val = (unsigned short) pci_read_config(osdev->dip, where, 2);

	return PCIBIOS_SUCCESSFUL;
}


int pci_read_config_dword(oss_device_t *osdev, offset_t where, 
		unsigned int *val) {
	*val = (unsigned int) pci_read_config(osdev->dip, where, 4);

	return PCIBIOS_SUCCESSFUL;
}


int pci_write_config_byte(oss_device_t * osdev, offset_t where,
		unsigned char val) {
	pci_write_config(osdev->dip, where, val, 1);

	return PCIBIOS_SUCCESSFUL;
}


int pci_write_config_word(oss_device_t * osdev, offset_t where,
		unsigned short val) {
	pci_write_config(osdev->dip, where, val, 2);

	return PCIBIOS_SUCCESSFUL;
}


int pci_write_config_dword(oss_device_t * osdev, offset_t where,
		unsigned int val) {
	pci_write_config(osdev->dip, where, val, 4);

	return PCIBIOS_SUCCESSFUL;
}


void oss_contig_free(void *p, unsigned long sz) {
	/*
	   cmn_err(0, "Not implemented: oss_contig_free(), (driver is leaking)\n");
	   */
	/* TODO */
}


void *oss_map_pci_mem(oss_device_t *osdev, int nr, oss_native_word paddr, 
		int psize) {
	return vm_map_phys(SELF, paddr, (size_t) psize);
}


int msg_unknown(int log, message *m_ptr) {
#if DEBUG_MESSAGES
	log = 1;
#endif
	if (log) cmn_err(0, "received unknown message, type: 0x%x\n", m_ptr->m_type);

	return EIO;
}


int msg_sig_handler(void) {
#if DEBUG_MESSAGES
	cmn_err(0, "received '%s'\n", "process status change");
#endif
	cmn_err(0, "Sound off\n");
	return 0;
}


int msg_update(void) {
#if DEBUG_MESSAGES
	cmn_err(0, "msg_update()");
#endif

	return 0;
}


int msg_signal(void) {
#if DEBUG_MESSAGES
	cmn_err(0, "msg_signal()");
#endif

	return 0;
}


int msg_interrupt(int intr_num) {
	osscore_intr_t *intr;
	int i;
	int retval;

#if DEBUG_MESSAGES
	cmn_err(0, "received '%s'\n", "interrupt");
#endif

	for (i = 0; i < nintrs; i++) {
		intr = &intrs[i];
		if (intr->top) {
			intr->top(intr->osdev);
		}
		if (intr->bottom) {
			intr->bottom(intr->osdev);
		}
		if ((retval = sys_irqenable(&(intr->irq_hook_id))) != OK) {
			cmn_err(CE_WARN, "ossintr: cannot enable irq: %d\n", retval);
			return retval;
		}
	}

	return 0;
}


int msg_cancel(message *m_ptr) {
#if DEBUG_MESSAGES
	cmn_err(0, "received '%s'\n", "cancel");
#endif
	return 0;
}


int msg_status(message *m_ptr) {
#if DEBUG_MESSAGES
	cmn_err(0, "received '%s'\n", "status");
#endif

	return 0;
}


int msg_write(message *m_ptr) {
	int retval;
	int minor = m_ptr->DEVICE;

#if DEBUG_MESSAGES
	cmn_err(0, "received '%s', minor: %d\n", "write", minor);
#endif

	uio_bufs[minor].uio_resid = m_ptr->COUNT;
	uio_bufs[minor].revive_proc_nr = m_ptr->IO_ENDPT;
	uio_bufs[minor].revive_grant = (cp_grant_id_t) m_ptr->ADDRESS;
	uio_bufs[minor].revive_status = 0;

	if (minor < 0 || minor >= OSS_MAX_CDEVS) {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, EIO, 
				uio_bufs[minor].revive_grant, "revive, after write, with error");
		return EIO;
	}

	retval = oss_write(minor, &uio_bufs[minor]);

	if (retval >= 0) {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, retval, 
				uio_bufs[minor].revive_grant, "revive, after write");
		return OK;
	}
	else {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, EIO, 
				uio_bufs[minor].revive_grant, "revive, after write, with error");
		return retval;
	}
}


int msg_read(message *m_ptr) {
	int retval;
	int minor = m_ptr->DEVICE;

#if DEBUG_MESSAGES
	cmn_err(0, "received '%s', minor: %d\n", "read", minor);
#endif

	uio_bufs[minor].uio_resid = m_ptr->COUNT;
	uio_bufs[minor].revive_proc_nr = m_ptr->IO_ENDPT;
	uio_bufs[minor].revive_grant = (cp_grant_id_t) m_ptr->ADDRESS;
	uio_bufs[minor].revive_status = 0;

	if (minor < 0 || minor >= OSS_MAX_CDEVS) {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, EIO, 
				uio_bufs[minor].revive_grant, "revive, after read, with error");
		return EIO;
	}

	retval = oss_read(m_ptr->DEVICE, &uio_bufs[minor]);

	if (retval >= 0)  {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, retval, 
				uio_bufs[minor].revive_grant, "revive, after read");
		return OK;
	}
	else {
		reply_with_grant(DEV_REVIVE, m_ptr->m_source, 
				uio_bufs[minor].revive_proc_nr, EIO, 
				uio_bufs[minor].revive_grant, "revive, after read, with error");
		return retval;
	}
}


int ioctl_length(int io_request) {
	io_request >>=16;
	return io_request & _IOCPARM_MASK;
}


int msg_ioctl(message *m_ptr) {
	int minor = m_ptr->DEVICE;
	int len = ioctl_length(m_ptr->REQUEST);
	int retval = 0;

#if DEBUG_MESSAGES
	cmn_err(0, "received '%s', minor: %d, len: %d\n", "ioctl", minor, len);
#endif

	memset(io_ctl_buf, 0, _IOCPARM_MASK);

	if (m_ptr->REQUEST & _IOC_IN) { /* if there is data for us, copy it */
		if(sys_safecopyfrom(m_ptr->IO_ENDPT,
					(vir_bytes)m_ptr->ADDRESS, 0,
					(vir_bytes)io_ctl_buf, len, D) != OK) {
			cmn_err(CE_WARN, "%s:%d: safecopyfrom failed\n", __FILE__, __LINE__);
		}
	}

	retval = oss_ioctl(minor, m_ptr->REQUEST, io_ctl_buf);

	if (retval == 0 && m_ptr->REQUEST & _IOC_OUT) { 
		/* copy back data, if requested */
		if(sys_safecopyto(m_ptr->IO_ENDPT, (vir_bytes)m_ptr->ADDRESS, 0,
					(vir_bytes)io_ctl_buf, len, D) != OK) {
			cmn_err(0, "%s:%d: safecopyto failed\n", __FILE__, __LINE__);
		}
	}

	return retval == 0 ? OK : EIO;
}


static void destroy_uio_buf(uio_t *uio_buf) {
	memset(uio_buf, '\0', sizeof(uio_t));
}


int msg_close(int minor) {
#if DEBUG_MESSAGES
	cmn_err(0, "received '%s', minor: %d\n", "close device", minor);
#endif

	if (minor < 0 || minor >= OSS_MAX_CDEVS) {
		return EIO;
	}
	destroy_uio_buf(&uio_bufs[minor]);

	return (oss_close(minor) == 0) ? OK : EIO;
}


int msg_open(int minor, int flags, int proc_nr) {
	int retval;
#if DEBUG_MESSAGES
	cmn_err(0, "received '%s', minor: %d, flags: 0x%x, proc_nr: %d\n", 
			"open device", minor, flags, proc_nr);
#endif
	if (minor < 0 || minor >= OSS_MAX_CDEVS) {
		return -1;
	}
	retval = oss_open(minor, flags, proc_nr);

	return retval == 0 ? OK : EIO;
}


int wait_for_interrupt(void) {
	int result = -1;
	message mess;

#if DEBUG_MESSAGES
	cmn_err(0, "waiting for '%s'\n", "interrupt");
#endif

	receive(ANY, &mess);

	if(is_notify(mess.m_type)) {
	   switch(_ENDPOINT_P(mess.m_source)) {
		case HARDWARE:
			result = msg_interrupt(mess.NOTIFY_ARG);
			break;
		case CLOCK:
			result = 0;
			expire_timers();
			break;
		default:
			result = add_to_message_queue(mess);
	  }
	} else {
		result = add_to_message_queue(mess);
	}
	return result;
}


void detect(char *module_name) {
	int i;
	module_t module;
	oss_device_t *osdev;

	for (i= 0; modules[i].name != NULL; i++) {
		module = modules[i];
		if (strncmp(module.name, module_name, strlen(module_name)) == 0) {
			if ((osdev = osdev_create (NULL, *(module.type),
							ndevs, module.name, NULL)) == NULL)
			{
				cmn_err(0, "Cannot create a device for %s\n", module.name);
				return;
			}

			if (!module.attach(osdev)) {
				cmn_err(0, "Cannot attach device for %s\n", module.name);
				return;
			}

			device_list[ndevs] = osdev;
			ndevs++;

			cmn_err(0, "Detected (virtual) device using module %s\n", module.name);
		}
	}
}


int device_matches_module(dev_info_t *device, module_t module) {
	int i;
	device_id_t *id_table = module.id_table;


	for (i = 0; id_table[i].vendor != 0; i++) {
		if (device->v_id == id_table[i].vendor &&
				device->d_id == id_table[i].device) {
			return 1;
		}
	}

	return 0;
}


int osspci_probe(dev_info_t *device) {
	int i;
	oss_device_t *osdev;
	module_t module;
	int result;

	for (i = 0; i < ndevs; i++) {
		if (device->v_id == minix_devices[i]->v_id &&
				device->d_id == minix_devices[i]->d_id) {
			cmn_err(0, "Device already detected\n");
			return -1;
		}
	}
	if (ndevs >= 16) {
		cmn_err(0, "Too many instances\n");
		return -1;
	}

	for (i = 0; modules[i].name != NULL; i++) {
		module = modules[i];
		if (device_matches_module(device, module)) {
			if ((osdev = osdev_create(device, *(module.type), ndevs, 
							module.name, NULL)) == NULL)
			{
				cmn_err(0, "Cannot create a device\n");
				return -1;
			}

			if ((result = pci_reserve_ok(device->devind)) != OK) {
				cmn_err(0, "Cannot reserve device: %d\n", result);
				return -1;
			}

			if (!module.attach(osdev)) {
				cmn_err(0, "Cannot attach device\n");
				return -1;
			}

			minix_devices[ndevs] = device;
			device_list[ndevs] = osdev;
			ndevs++;

			cmn_err(0, "Detected device: 0x%x, 0x%x using module %s\n", 
					device->v_id, device->d_id, module.name);
			return 0;
		}
	}
	return -1;
}


void print_device(dev_info_t *device) {
	cmn_err(0, "\tvendor: 0x%x, device: 0x%x\n", device->v_id, device->d_id);
	cmn_err(0, "\tname: %s\n", device->name == NULL ? "no name" : device->name);
	cmn_err(0, "\tslot name: %s\n", device->slot_name);
	cmn_err(0, "\tbase address: 0x%x\n", device->base_address);
	cmn_err(0, "\tirq: %d\n", device->irq);
}


void detect_hardware(void) {
	int retval;
	int devind;
	u16_t v_id, d_id;
	dev_info_t *device;

	pci_init();

	retval = pci_first_dev(&devind, &v_id, &d_id);
	if (retval > 0) {
		do {
			device = malloc(sizeof(dev_info_t));
			device->v_id = v_id;
			device->d_id = d_id;
			device->devind = devind;
			device->name = pci_dev_name(v_id, d_id);
			device->slot_name = pci_slot_name(devind);
			device->base_address = pci_attr_r32(devind, PCI_BAR) & 0xfffffffe;
			device->irq = pci_attr_r8(devind, PCI_ILR);
			cmn_err(0, "Trying to detect device:\n");
			print_device(device);
			if ((osspci_probe(device)) != 0) {
				cmn_err(0, "Detecting device 0x%x, 0x%x failed\n", device->v_id, 
						device->d_id);
			}
		}
		while((retval = pci_next_dev(&devind, &v_id, &d_id)));
	}
	else {
		cmn_err(0, "No devices detected\n");
	}
}


/* key and value will point into argument, argument will be modified */
void toKeyValue(char *argument, char **key, char **value) {
	int arglen=strlen(argument);
	int i;

	*key = argument;
	for (i = 0; i < arglen; i++) {
		if (argument[i] == '=') {
			argument[i] = '\0';
			*value = argument + i + 1;
			break;
		}
	}
	if (i == arglen - 1) {
		cmn_err(0, "ERROR!!!\n");
	}
}


void process_argument(char *arg) {
	char *key;
	char *value;
	char argument[1024];
	option_t option;
	int i;
	int valueInt;


	strncpy(argument, arg, 1024);
	argument[1023]='\0';

	toKeyValue(argument, &key, &value);

	for (i = 0; option_table[i].name != NULL; i++) {
		option = option_table[i];
		if (strcmp(option.name, key) == 0) {
			valueInt = atoi(value);
			cmn_err(0, "setting option: %s=%d\n", option.name, valueInt);
			*option.var = valueInt;
		}
	}
}


void process_arguments(int argc, char **argv) {
	char *argument;

	if (argc == 2) {
		argument = strtok(argv[1], ",");
		while (argument != NULL) {
			process_argument(argument);
			argument = strtok(NULL, ",");
		}
	}
}


int main(int argc, char **argv)
{
	message mess;
	int result = 0;
	int source;
	int io_endpt;
	struct fssignon device;

	cmn_err(0, "Sound on\n");

	oss_hz = sys_hz();
	process_arguments(argc, argv);
	soundcard_attach();
	detect_hardware();
	/*detect("oss_imux");*/ /* doesn't work and is obsolete, vmix takes over
							   all functionality and works better */
	/*detect("audioloop"); */

	device.dev = makedev(oss_major, 0);
	device.style = 0;

	if(svrctl(FSSIGNON, (void *) &device) == -1) {
		cmn_err(0, "OSS: couldn't register as device driver");
		exit(0);
	}

	while(1) {
		if (messages_in_queue()) {
			mess = get_message_from_queue();
		}
		else {
			receive(ANY, &mess);
		}
		source = mess.m_source;
		io_endpt = mess.IO_ENDPT;

		if(is_notify(mess.m_type)) {
		   switch(_ENDPOINT_P(mess.m_source)) {
			case RS_PROC_NR:
				notify(RS_PROC_NR);
				result = OK;
#if DEBUG_MESSAGES
					cmn_err(0, "received 'ping'\nreplied 'ping'\n");
#endif
				break;
			case CLOCK:
				expire_timers();
				result = OK;
				break;
			case HARDWARE:
				result = msg_interrupt(mess.NOTIFY_ARG);
				break;
			case SYSTEM:
				result = msg_signal();
				break;
			case DS_PROC_NR:
				result = msg_update();
				break;
			case PM_PROC_NR:
				result = msg_sig_handler();
				break;

			default:
				cmn_err(0, "unrecognized notify()\n");
				continue;
				break;
		    }
		} else {
		   switch(mess.m_type) {
			case DEV_OPEN:
				result = msg_open(mess.DEVICE, mess.COUNT, io_endpt);
				reply(DEV_REVIVE, source, io_endpt, result, "revive, after open");
				break;
			case DEV_CLOSE:
				result = msg_close(mess.DEVICE);
				reply(DEV_CLOSE_REPL, source, io_endpt, result, "device close");
				break;
			case DEV_IOCTL_S:
				result = msg_ioctl(&mess);
				reply_with_grant(DEV_REVIVE, source, io_endpt, result, 
						(unsigned) mess.IO_GRANT, "revive, after ioctl");
				break;
			case DEV_READ_S:
				result = msg_read(&mess);
				break;
			case DEV_WRITE_S:
				result = msg_write(&mess);
				break;
			case DEV_STATUS:
				result = msg_status(&mess);
				break;
			case CANCEL:
				result = msg_cancel(&mess);
				break;
			case DEV_REOPEN:
				result = msg_open(mess.DEVICE, mess.COUNT, io_endpt);
				reply(DEV_REOPEN_REPL, source, io_endpt, result, "device reopened");
				break;
			default:
				cmn_err(0, "unknown message: 0x%x\n", mess.m_type);
				result = msg_unknown(LOG_UNKNOWN_MESSAGE, &mess);
				break;
		   }
		}
		if (result != OK) {
			cmn_err(0, "ERROR, the last action failed with return code: %d\n",
					result);
			cmn_err(0, "mess.m_type: 0x%x\n", mess.m_type);
		}
	} /* end receive messages loop */
	return 0;
}
